#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

"""
Character-based functions for enriching English language data.

Importable functions include:

* add_characters
* add_contractions
* add_fat_thumbs
* add_leet
* add_macbook_keyboard
* add_whitespace
* remove_characters
* remove_contractions
* remove_punctuation
* remove_whitespace
* swap_chars
"""

import collections
import json
from pkg_resources import resource_string
from string import ascii_letters, punctuation
import typing

from scipy import random


LEETMAP = collections.OrderedDict(
    [
        ("anned", "&"),
        ("and", "&"),
        ("what", "wat"),
        ("are", "r"),
        ("ate", "8"),
        ("at", "@"),
        ("one", "1"),
        ("you", "u"),
        ("t", "7"),
        ("o", "0"),
        ("e", "3"),
        ("l", "1"),
    ]
)

CONTRACT = json.loads(
    resource_string("niacin", "data/contractions.json").decode("utf-8")
)
EXPAND = {v: k for k, v in CONTRACT.items()}
NEIGHBORS = json.loads(resource_string("niacin", "data/neighbors.json").decode("utf-8"))


def _sub_chars(string: str, probability: float, mapping: typing.Mapping) -> str:
    """Replace substrings with a given probability.

    Given a mapping, search string one by one for keys and replace with
    the appropriate value, with some probability. If your keys are not mutually
    exclusive (e.g. some part of them overlaps), the order in which they appear
    in the mapping becomes important.

    Args:
        string: text
        probability: probability of replacing a group of characters
        mapping: map of substring -> replacement

    Returns:
        enriched text
    """
    for pattern, sub in mapping.items():
        index = 0
        while 0 <= index < len(string):
            index = string.lower().find(pattern, index)
            if index < 0:
                break
            elif random.binomial(1, probability):
                string = string[:index] + sub + string[index + len(pattern) :]
                index += len(sub)
            else:
                index += len(pattern)
    return string


def add_fat_thumbs(string: str, p: float = 0.01) -> str:
    """Replace characters with QWERTY neighbors.

    One source of typographic mistakes comes from pressing a nearby key
    on a keyboard (or on a touchscreen). With probability p, replace each
    character is a string with one from a set of its neighbors. The
    replacement is chosen using ``random.choice``.

    Args:
        string: text
        p: probability of replacing a character

    Returns:
        enriched text
    """
    for index, char in enumerate(string):
        if char in NEIGHBORS and random.binomial(1, p):
            new_char = random.choice(NEIGHBORS[char])
            string = string[:index] + new_char + string[index + 1 :]
    return string


def add_characters(string: str, p: float = 0.01) -> str:
    """Insert individual characters with probability p.

    These are chosen randomly from the ascii alphabet (including
    both upper and lower cases).

    Args:
        string: text
        p: probability of removing a character

    Returns:
        enriched text
    """
    for index in reversed(range(len(string))):
        if random.binomial(1, p):
            new_char = random.choice(list(ascii_letters))
            string = string[:index] + new_char + string[index:]
    return string


def add_contractions(string: str, p: float = 0.5) -> str:
    """Replace common word pairs with their contraction.

    This is done even when the contraction introduces ambiguity, as this is
    seen as preserving the semantics (arXiv:1812.04718_).

    Args:
        string: text
        p: probability of a word pair being replaced

    Returns:
        enriched text

    .. _arXiv:1812.04718 : https://arxiv.org/abs/1812.04718
    """
    return _sub_chars(string, probability=p, mapping=CONTRACT)


def remove_contractions(string: str, p: float = 0.5) -> str:
    """Expand a contraction into individual tokens.

    See (arXiv:1812.04718_).

    Args:
        string: text
        p: probability of a word pair being replaced

    Returns:
        enriched text

    .. _arXiv:1812.04718 : https://arxiv.org/abs/1812.04718
    """
    return _sub_chars(string, probability=p, mapping=EXPAND)


def add_leet(string: str, p: float = 0.2) -> str:
    """Replace character groups with visually or aurally similar ones.

    Character groups given in ``LEETMAP.keys()`` are searched for in
    priority (roughly from largest to smallest), and are replaced with
    some associated value with probability p. E.g.:

    | "Hello, you are banned"
    | "Hello, you are b&"
    | "Hello, you r b&"
    | "Hello, u r b&"
    | "H3110, u r b&"

    Args:
        string: text
        p: condtional probability of replacing a character group

    Returns:
        enriched text
    """
    return _sub_chars(string, probability=p, mapping=LEETMAP)


def add_macbook_keyboard(string: str, p: float = 0.1) -> str:
    """Repeats or removes each character with probability p.

    Bad keyboards can be a common source of typographical errors by
    repeating characters or by omitting them, e.g. because the individual
    keys get stuck. With probability p, we modify a character, with a
    50/50 chance of either removing it, or repeating it twice.

    Args:
        string: text
        p: probability of changing letter count

    Returns:
        enriched text
    """
    for index in reversed(range(len(string))):
        if random.binomial(1, p):
            count = random.choice([0, 2])
            string = string[:index] + string[index]*count + string[index+1:]
    return string


def add_whitespace(string: str, p: float = 0.01) -> str:
    """Add a spacebar character with probability p.

    Extraneous whitespace, especially when it occurs in the middle of an
    important word, can be reduce the effectiveness of models which depend
    on word tokenizers as part of the data pipeline.

    Args:
        string: text
        p: probability of adding a space character

    Returns:
        enriched text
    """
    space = " "
    for index in range(len(string), -1, -1):
        if random.binomial(1, p):
            string = string[:index] + space + string[index:]
    return string


def remove_characters(string: str, p: float = 0.01) -> str:
    """Remove individual characters with probability p.

    Args:
        string: text
        p: probability of removing a character

    Returns:
        enriched text
    """
    for index in reversed(range(len(string))):
        if random.binomial(1, p):
            string = string[:index] + string[index + 1 :]
    return string


def remove_punctuation(string: str, p: float = 0.25) -> str:
    """Remove punctuation with probability p.

    The removal of punctuation is a common data cleaning step for fast but
    high bias models and data processing algorithms. When that punctuation
    occurs in the middle of the word (e.g. indicating possessiveness), its
    removal may change the semantics of the string.

    Args:
        string: text
        p: probability of removing punctuation

    Returns:
        enriched text
    """
    mapping = {k: "" for k in punctuation}
    return _sub_chars(string, probability=p, mapping=mapping)


def remove_whitespace(string: str, p: float = 0.1) -> str:
    """Remove a spacebar character with probability p.

    Selective removal of whitespace can be reduce the effectiveness of word-
    based models, or those which depend on word tokenizers as part of the
    data pipeline.

    Args:
        string: text
        p: probability of removing a space character

    Returns:
        enriched text
    """
    mapping = {" ": ""}
    return _sub_chars(string, probability=p, mapping=mapping)


def swap_chars(string: str, p: float = 0.05) -> str:
    """Swap adjacent characters.

    With probability p, swap two adjacent characters in a string. No
    character gets swapped more than once, so cannot end up in any locations
    that are not adjacent to its starting position.

    .. note::
        to keep the interface consistent, niacin's implementation acts on
        a probability p, applied n-1 times, where n is the total number of
        characters in the string. The implementation in noisemix_ (called
        ``flip_chars``) chooses two letters at random and exchanges their
        positions, exactly once per string.

    Args:
        string: text
        p: probability of swapping two characters

    Returns:
        enriched text

    .. _noisemix : https://github.com/noisemix/noisemix
    """
    chars = list(string)
    index = 0
    while index < len(chars) - 1:
        if random.binomial(1, p):
            chars[index], chars[index + 1] = chars[index + 1], chars[index]
            index += 2
        else:
            index += 1
    return "".join(chars)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Classes for compatibility with PyTorch data loaders. Includes:

* MemoryTextDataset
* FileTextDataset
* DirectoryTextDataset
"""

import collections
import os
from pathlib import Path
import typing as t

from nltk import WordPunctTokenizer
import pandas as pd
import torch
from torch.utils.data import Dataset
from torchtext.vocab import Vocab


class TextDatasetMixin:

    _data: t.Sequence
    _tokenizer: t.Callable
    _transforms: t.Iterable[t.Callable]
    _vocab: Vocab

    def _build_vocab(self) -> Vocab:
        counter: collections.Counter = collections.Counter()
        # older versions of torchtext require special symbols to be in the
        # vocabulary
        counter.update(["<unk>", "<pad>"])
        for doc in self._data:
            for token in self._tokenizer(doc):
                counter.update([token])
        return Vocab(counter)

    def _tokenize(self, text: str) -> t.List[str]:
        return self._tokenizer(text)

    def _transform(self, text: str) -> str:
        for transform in self._transforms:
            text = transform(text)
        return text

    def _vectorize(self, tokens: t.List[str]) -> torch.Tensor:
        return torch.tensor([self._vocab[token] for token in tokens])


class MemoryTextDataset(Dataset, TextDatasetMixin):
    """A text dataset for data that is already in memory. Accepts a list of
    text inputs and a list of labels.

    Args:
        data: list of text inputs
        labels: list of text labels
        transforms: either a list of transformation functions, or an Augmenter
            class
        tokenizer: a function that receives text and returns tokens
        vocab: a PyTorch Vocab object
    """

    def __init__(
        self,
        data: t.Sequence,
        labels: t.Sequence,
        transforms: t.Iterable[t.Callable] = None,
        tokenizer: t.Callable = None,
        vocab: Vocab = None,
    ):
        self._data = data
        self._labels = labels
        if transforms is None:
            self._transforms = []
        else:
            self._transforms = transforms
        if tokenizer is None:
            self._tokenizer = WordPunctTokenizer().tokenize  # type: ignore
        else:
            self._tokenizer = tokenizer  # type: ignore
        if vocab is None:
            self._vocab = self._build_vocab()
        else:
            self._vocab = vocab

    def __len__(self) -> int:
        return len(self._labels)

    def __getitem__(self, index: int) -> t.Tuple[torch.Tensor, t.Union[int, float]]:
        label = self._labels[index]
        data = self._data[index]
        data = self._transform(data)
        tokens = self._tokenize(data)
        vector = self._vectorize(tokens)
        return vector, label


class FileTextDataset(MemoryTextDataset):
    """A text dataset for data that is in a single file on disk. The reader
    assumes that the labels will be in the first column, and all text data
    will be in subsequent columns.

    Args:
        datafile: location of file on disk
        sep: delimiter between columns of data in file
        transforms: either a list of transformation functions, or an Augmenter
            class
        tokenizer: a function that receives text and returns tokens
        vocab: a PyTorch Vocab object
    """

    def __init__(
        self,
        datafile: str,
        sep: str = ",",
        transforms: t.Iterable[t.Callable] = None,
        tokenizer: t.Callable = None,
        vocab: Vocab = None,
    ):
        df = pd.read_table(datafile, sep=sep)
        self._data = df.iloc[:, 1:].apply(lambda s: s.str.cat(sep=" "), axis=1).tolist()
        self._labels = df.iloc[:, 0].tolist()
        if transforms is None:
            self._transforms = []
        else:
            self._transforms = transforms
        if tokenizer is None:
            self._tokenizer = WordPunctTokenizer().tokenize  # type: ignore
        else:
            self._tokenizer = tokenizer  # type: ignore
        if vocab is None:
            self._vocab = self._build_vocab()
        else:
            self._vocab = vocab


class DirectoryTextDataset(Dataset, TextDatasetMixin):
    """A text dataset for data that is spread across multiple files.

    Assumes two separate directories exist: one with text data, and one with
    labels, in separate files with the same names (but not necessarily the same
    extensions). E.g.

    * data/
        * review_one.txt
        * review_two.txt
        * ...
    * labels/
        * review_one.label
        * review_two.label
        * ...

    Args:
        data_dir: folder with data files
        labels_dir: folder with label files
        transforms: either a list of transformation functions, or an Augmenter
            class
        tokenizer: a function that receives text and returns tokens
        vocab: a PyTorch Vocab object
    """

    def __init__(
        self,
        data_dir: str,
        labels_dir: str,
        transforms: t.Iterable[t.Callable] = None,
        tokenizer: t.Callable = None,
        vocab: Vocab = None,
    ):
        self._data_dir = Path(data_dir)
        self._data = sorted(os.listdir(data_dir))
        self._labels_dir = Path(labels_dir)
        self._labels = sorted(os.listdir(labels_dir))
        if transforms is None:
            self._transforms = []
        else:
            self._transforms = transforms
        if tokenizer is None:
            self._tokenizer = WordPunctTokenizer().tokenize  # type: ignore
        else:
            self._tokenizer = tokenizer  # type: ignore
        if vocab is None:
            self._vocab = self._build_vocab()
        else:
            self._vocab = vocab

    def __len__(self) -> int:
        return len(self._labels)

    def __getitem__(self, index: int) -> t.Tuple[torch.Tensor, torch.Tensor]:
        filename = self._labels[index]
        label_path = self._labels_dir / filename
        with open(label_path) as f:
            label = torch.tensor(float(f.read()))
        data_path = self._data_dir / filename
        with open(data_path) as f:
            data = f.read()
        data = self._transform(data)
        tokens = self._tokenize(data)
        vector = self._vectorize(tokens)
        return vector, label

    def _build_vocab(self) -> Vocab:
        counter: collections.Counter = collections.Counter()
        # older versions of torchtext require special symbols to be in the
        # vocabulary
        counter.update(["<unk>", "<pad>"])
        for filename in self._data:
            fp = self._data_dir / filename
            with open(fp) as f:
                doc = f.read()
            for token in self._tokenizer(doc):
                counter.update([token])
        return Vocab(counter)



#!/usr/bin/env python
# -*- coding: utf-8

"""Time domain transforms
"""


import numpy as np
from numpy import random
from scipy import interpolate



def add_slope_trend(x: np.ndarray, p: float=0.01, m: float=0.1) -> np.ndarray:
    r"""Add linear trend, with probability p, and magnitude m*std(x).

    The probability refers to the entire trend -- either it is added, or the
    original array is left alone. The direction of the trend is chosen randomly
    with probability=0.5.

    Args:
        x: sequence
        p: per-sequence probability of applying trend
        m: magnitude of trend

    Returns:
        enriched sequence

    Examples:
        >>> x = np.sin(np.linspace(0, 6*np.pi, 100))
        |
        +1  -\\               -\\               --\
        |  /  \\             /  \\             /   \
        | /    \             /    \            /    \
        | /     \           /     \           /     \
        |/      \           /      \          /      \
        |/       \         /       \         /       \
        +--------\--------/--------\---------/-------\---------+--
        +0        \       /         \       /         \     +99
        |         \      /          \      /           \      /
        |          \     /           \     /           \     /
        |          \    /            \    /             \    /
        |           \   /             \\  /             \\  /
        -1           --/               \-/               \-/
        |
        >>> ts.add_slope_trend(x, 1.0, 1.0)
        |
        +1.52476                                --\\
        |                     --\              /   \
        |   --\              // \\            //    \
        |  // \\             /    \           /      \
        | /    \            /     \          /       \         *
        | /     \           /      \         /        \       /
        |/      \          /       \\       /          \      /
        |/       \        /         \      /           \     /
        +---------\-------/----------\-----/------------\----/-+--
        +0        \      /           \    /             \\ /+99
        |          \    //            \\ //              --/
        |           \   /              --/
        -0.82       \--/
        |
    """
    if random.binomial(1, p):
        s = np.nanstd(x)
        sign = (random.binomial(1, 0.5, 1) * 2 - 1).item()
        x = x + np.linspace(0, sign*m*s, len(x))
    return x


def add_spike(x: np.ndarray, p: float=0.01, m: float=1.0) -> np.ndarray:
    r"""At each array entry, add a spike with probability p and magnitude
    m*std(x).

    The direction of the spike (up or down) is determined randomly with
    probability=0.5.

    Args:
        x: sequence
        p: per-entry probability of adding spike
        m: magnitude of spike

    Returns:
        enriched sequence

    Examples:
        >>> x = np.sin(np.linspace(0, 6*np.pi, 100))
        |
        +1  -\\               -\\               --\
        |  /  \\             /  \\             /   \
        | /    \             /    \            /    \
        | /     \           /     \           /     \
        |/      \           /      \          /      \
        |/       \         /       \         /       \
        +--------\--------/--------\---------/-------\---------+--
        +0        \       /         \       /         \     +99
        |         \      /          \      /           \      /
        |          \     /           \     /           \     /
        |          \    /            \    /             \    /
        |           \   /             \\  /             \\  /
        -1           --/               \-/               \-/
        |
        >>> ts.add_spike(x, 0.1, 1.0)
        |
        +1.61319              |
        | |                  ||
        ||| -|\              |--\               |-\
        ||// |\\             |  |            | /|  \
        ||   | \\           /   | \          |/ |   |
        ||   |  \           /   || \         |/ |   |\
        +--------\\-------||-----|-\\--------|------|\|--------+--
        +0        \      /||        \   |  //       | |     +99
        |          \     / |         | ||  /          |\\    /
        |           \  //            | ||//           ||\| ///
        |            --/             |||-/             | |-/
        |                            \||                 |
        -1.64856                      ||                 |
        |
    """
    s = np.nanstd(x)
    ps = random.binomial(1, p, len(x))
    signs = random.binomial(1, 0.5, len(x)) * 2 - 1
    x = x + ps * signs * (m * s)
    return x


def add_step_trend(x: np.ndarray, p: float=0.01, m: float=0.1) -> np.ndarray:
    r"""Add a stepwise trend, where each entry in the timeseries has p
    probability of a stepwise change of magnitude m*std(x).

    The direction of the stepwise trend is chosen with probability=0.5 for
    the entire trend (e.g. if the first step is upward, then every subsequent
    step is also upward).

    Args:
        x: sequence
        p: per-entry probability of step change
        m: magnitude of step change

    Returns:
        enriched sequence

    Examples:
        >>> x = np.sin(np.linspace(0, 6*np.pi, 100))
        |
        +1  -\\               -\\               --\
        |  /  \\             /  \\             /   \
        | /    \             /    \            /    \
        | /     \           /     \           /     \
        |/      \           /      \          /      \
        |/       \         /       \         /       \
        +--------\--------/--------\---------/-------\---------+--
        +0        \       /         \       /         \     +99
        |         \      /          \      /           \      /
        |          \     /           \     /           \     /
        |          \    /            \    /             \    /
        |           \   /             \\  /             \\  /
        -1           --/               \-/               \-/
        |
        >>> ts.add_step_trend(x, 0.05, 0.5)
        |
        +1 /--\              /--\
        | /    \\           //   |
        |/      \          //     \
        +--------\\-------/-------\\---------------------------+--
        +0        \\     /         \\         \ --\         +99
        |          \\   //          \\        //   \\
        |           \--/             |       /      \\
        |                             \   \//        \
        |                              --///          \\
        |                                              \\
        |                                               \\
        |                                                 -\ \/*
        -3.02032                                           ////
        |
    """
    s = np.nanstd(x)
    sign = (random.binomial(1, 0.5, 1) * 2 - 1).item()
    ps = random.binomial(1, p, len(x))
    x = x + np.cumsum(ps * (sign * m * s))
    return x


def add_warp(x: np.ndarray, p: float=0.01, m: float=0.1, interp_method: str = 'linear') -> np.ndarray:
    r"""Warp the distances between points in a timeseries.

    With probability p, upsample a timeseries by a scale of m*len(x). Then,
    randomly len(x) items from upsampled timeseries, where choice is uniform.

    Args:
        x: sequence
        p: per-sequence probability of warping
        m: magnitude of warp
        interp_method: scipy interp1d kind: can be ‘linear’, ‘nearest’,
            ‘nearest-up’, ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’, ‘previous’,
            or ‘next’. More details at scipy.interpolate.interp1d_.

    Returns:
        enriched sequence

    Examples:
        >>> x = np.sin(np.linspace(0, 6*np.pi, 100))
        |
        +1  -\\               -\\               --\
        |  /  \\             /  \\             /   \
        | /    \             /    \            /    \
        | /     \           /     \           /     \
        |/      \           /      \          /      \
        |/       \         /       \         /       \
        +--------\--------/--------\---------/-------\---------+--
        +0        \       /         \       /         \     +99
        |         \      /          \      /           \      /
        |          \     /           \     /           \     /
        |          \    /            \    /             \    /
        |           \   /             \\  /             \\  /
        -1           --/               \-/               \-/
        |
        >>> ts.add_warp(x, 1.0, 0.1)
        |
        +1 -\                ---\                 -\\
        | /  \|             |   --\              /  -\
        |-/   |             |     \             /     |
        ||    |            |      \\            |     |
        ||    -\          -|       |            |     |
        ||      |        /         |            |     |
        +-------|--------/---------|-----------/-------\------=+--
        +0      |       /           \         /        |    +99
        |        \     -/            \\       /        |     /
        |        \    |                \     /          |    /
        |         \   |                 \    /          |  //
        |          -  |                 \   |           |-/
        -1          -/                   -/-            --
        |

    .. _scipy.interpolate.interp1d: https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html
    """
    old_size = len(x)
    step = round(old_size * m)
    if step < 2:
        # can't warp with no extra space
        return x
    if random.binomial(1, p):
        stretch_size = old_size * step
        old_t = np.linspace(0, 1, old_size)
        stretch_t = np.linspace(0, 1, stretch_size)
        select = np.sort(random.choice(stretch_t, size=old_size, replace=False))
        f = interpolate.interp1d(old_t, x, kind=interp_method)
        x = f(select)
    return x


def crop_and_stretch(x: np.ndarray,  p: float=0.01, m: float=0.1, interp_method: str = 'linear') -> np.ndarray:
    r"""Crop a sequence and stretch remaining entries to be the original size.

    With probability p, crop x such that it has m*len(x) entries remaining.
    Then, stretch it back to size len(x) using given interpolation method
    (default method is linear).

    Args:
        x: sequence
        p: per-sequence probability of cropping
        m: magnitude of crop (larger m is fewer entries from x)
        interp_method: scipy interp1d kind: can be ‘linear’, ‘nearest’,
            ‘nearest-up’, ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’, ‘previous’,
            or ‘next’. More details at scipy.interpolate.interp1d_.

    Returns:
        enriched sequence

    Examples:
        >>> x = np.sin(np.linspace(0, 6*np.pi, 100))
        |
        +1  -\\               -\\               --\
        |  /  \\             /  \\             /   \
        | /    \             /    \            /    \
        | /     \           /     \           /     \
        |/      \           /      \          /      \
        |/       \         /       \         /       \
        +--------\--------/--------\---------/-------\---------+--
        +0        \       /         \       /         \     +99
        |         \      /          \      /           \      /
        |          \     /           \     /           \     /
        |          \    /            \    /             \    /
        |           \   /             \\  /             \\  /
        -1           --/               \-/               \-/
        |
        >>> ts.crop_and_stretch(x, 1.0, 0.5)
        |
        +1---\                               -----\
        |     \\                           ///     \\
        |      \\                         //        \\
        |       \\                       /           \\
        |         \                     /              \
        |          \                   //               \
        +-----------\-----------------//-----------------\-----+--
        +0           \               /                   \\ +99
        |            \\             /                     \\
        |              \           //                      \\
        |               \         //                         \
        |                \\     ///                           \*
        -1                \----//
        |

    .. _scipy.interpolate.interp1d: https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html
    """
    old_size = len(x)
    crop_size = round((1-m) * old_size)
    if (crop_size < 1) or (crop_size >= old_size):
        # if m << there might be no crop
        # if m >> the crop might be equal to length of sequence
        return x
    if random.binomial(1, p):
        start = random.randint(0, old_size-crop_size, 1).item()
        crop = x[start:start+crop_size]
        crop_t = np.linspace(0, 1, crop_size)
        new_t = np.linspace(0, 1, old_size)
        f = interpolate.interp1d(crop_t, crop, kind=interp_method)
        x = f(new_t)
    return x


def flip(x: np.ndarray, p: float=0.5, m=None) -> np.ndarray:
    r"""Flip sequence around origin with probability p.

    Args:
        x: sequence
        p: per-sequence probability of flipping
        m: ignored

    Returns:
        enriched sequence

    Examples:
        >>> x = np.sin(np.linspace(0, 6*np.pi, 100))
        |
        +1  -\\               -\\               --\
        |  /  \\             /  \\             /   \
        | /    \             /    \            /    \
        | /     \           /     \           /     \
        |/      \           /      \          /      \
        |/       \         /       \         /       \
        +--------\--------/--------\---------/-------\---------+--
        +0        \       /         \       /         \     +99
        |         \      /          \      /           \      /
        |          \     /           \     /           \     /
        |          \    /            \    /             \    /
        |           \   /             \\  /             \\  /
        -1           --/               \-/               \-/
        |
        >>> ts.flip(x, 1.0)
        |
        +1           --\               /-\               /-\
        |           /   \             //  \             //  \
        |          /    \            /    \             /    \
        |          /     \           /     \           /     \
        |         /      \          /      \           /      \
        |         /       \         /       \         /       \
        +--------/--------\--------/---------\-------/---------+--
        +0       /         \       /         \       /      +99
        |\      /           \      /          \      /
        | \     /           \     /           \     /
        | \    /             \    /            \    /
        |  \  //             \  //             \   /
        -1  -//               -//               --/
        |
    """
    if random.binomial(1, p):
        x = -x
    return x


def reverse(x: np.ndarray, p: float=0.5, m=None) -> np.ndarray:
    r"""Reverse order of sequence with probability p

    Args:
        x: sequence
        p: per-sequence probability of reversal
        m: ignored

    Returns:
        enriched sequence

    Examples:
        >>> x = np.sin(np.linspace(0, 6*np.pi, 100))
        |
        +1  -\\               -\\               --\
        |  /  \\             /  \\             /   \
        | /    \             /    \            /    \
        | /     \           /     \           /     \
        |/      \           /      \          /      \
        |/       \         /       \         /       \
        +--------\--------/--------\---------/-------\---------+--
        +0        \       /         \       /         \     +99
        |         \      /          \      /           \      /
        |          \     /           \     /           \     /
        |          \    /            \    /             \    /
        |           \   /             \\  /             \\  /
        -1           --/               \-/               \-/
        |
        >>> ts.reverse(x, 1.0)
        |
        +1           --\               /-\               /-\
        |           /   \             //  \             //  \
        |          /    \            /    \             /    \
        |          /     \           /     \           /     \
        |         /      \          /      \           /      \
        |         /       \         /       \         /       \
        +--------/--------\--------/---------\-------/---------+--
        +0       /         \       /         \       /      +99
        |\      /           \      /          \      /
        | \     /           \     /           \     /
        | \    /             \    /            \    /
        |  \  //             \  //             \   /
        -1  -//               -//               --/
    """
    if random.binomial(1, p):
        x = x[::-1]
    return x


#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

"""
Functions for enriching timeseries data.

Includes transformations which operate in the time domain, and ones which
operate in the frequency domain. Organization of this module and basic set
of functionality are inspired by Wen et al.'s review, "Timeseries data
Augmentation for Deep Learning" (arXiv:2002.12478_).

Importable functions include:

Time domain
-----------

* add_slope_trend
* add_spike
* add_step_trend
* add_warp
* crop_and_stretch
* flip
* reverse


Frequency domain
----------------

* add_discrete_phase_shift
* add_high_frequency_noise
* add_random_frequency_noise
* remove_random_frequency

.. _arXiv:2002.12478 : https://arxiv.org/abs/2002.12478
"""

from .freq import (add_discrete_phase_shifts, add_high_frequency_noise, add_random_frequency_noise, remove_random_frequency)
from .time import (add_slope_trend, add_spike, add_step_trend, add_warp, crop_and_stretch, flip, reverse)


#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

"""
Implementation of RandAugment algorithm
"""

from functools import partial
import typing as t

from numpy.random import default_rng


class RandAugment:
    """Implements RandAugment algorithm (randaugment_) as an iterator.
    RandAugment selects ``n`` functions at random from a sequence, and
    initializes them with magnitude ``m``.

    To use it, initialize this class with a list of transformation functions.
    Every time it is iterated over (e.g. in a for loop) it yields a random
    subset of those transformation functions.

    The original paper claims that m is on a scale from 0-10, but its listed
    experiments regularly use magnitudes in the 20s and 30s. We take this to
    mean that "10" was a typo, and the scale was meant to extend to 100.

    The defaults for m and n have been set to 1 and 10, respectively, for
    safety. Depending on your model size and task, you may achieve more
    accurate results with n ∈ {2, 3} and an m in the range [10, 20). By
    default, the transforms in a single sample will be returned in random
    order. If your transforms must occur in a logical sequence (e.g. swapping
    synonyms before removing random characters), set shuffle to False.

    Args:
        transforms: sequence of transformation functions
        m: magnitude of transformation, on a scale of 0-100
        n: number of transforms to select
        shuffle: return transforms in random order
        seed: seed to use for the random number generator

    .. _randaugment: https://arxiv.org/abs/1909.13719
    """

    # niacin functions use floats for magnitude
    _p: float = 0.0

    def __init__(
        self,
        transforms: t.List[t.Callable],
        m: int = 10,
        n: int = 1,
        shuffle: bool = True,
        seed: int = None
    ):
        self._transforms = list(transforms)
        self.n = n
        self.m = m
        self._shuffle = shuffle
        self._rng = default_rng(seed=seed)

    def __len__(self):
        return len(self._transforms)

    def __iter__(self):
        choices = [
            partial(fun, p=self._p) for fun in self._rng.choice(
                self._transforms, size=self._n, replace=False, shuffle=self._shuffle
            )
        ]
        return iter(choices)

    @property
    def n(self) -> int:
        """Size of sample - should be less than total available transforms
        """
        return self._n

    @n.setter
    def n(self, value: int):
        n_funs = len(self._transforms)
        if value > n_funs:
            msg = f"Sample size n={value} must be <= number of transforms={n_funs}"
            raise ValueError(msg)
        self._n = value

    @property
    def m(self) -> int:
        """Magnitude of transformation - should be a number between 0 and 100
        """
        return int(self._p * 100)

    @m.setter
    def m(self, value: int):
        self._p = min(max(value / 100, 0.0), 1.0)


#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

"""
Functions for enriching English language data.

Includes transformations which operate on characters, words, and whole
sentences. Importable functions include:

Character-based
---------------

* add_characters
* add_contractions
* add_fat_thumbs
* add_leet
* add_macbook_keyboard
* add_whitespace
* remove_characters
* remove_contractions
* remove_punctuation
* remove_whitespace
* swap_chars

Word-based
----------

* add_hypernyms
* add_hyponyms
* add_misspelling
* add_parens
* add_synonyms
* remove_articles
* swap_words

Sentence-based
--------------

* add_applause
* add_backtranslation
* add_bytes
* add_love
"""

from .char import (
    add_characters,
    add_contractions,
    remove_contractions,
    add_fat_thumbs,
    add_leet,
    add_macbook_keyboard,
    add_whitespace,
    remove_characters,
    remove_punctuation,
    remove_whitespace,
    swap_chars,
)
from .sentence import add_applause, add_backtranslation, add_bytes, add_love
from .word import (
    add_hypernyms,
    add_hyponyms,
    add_misspelling,
    add_parens,
    add_synonyms,
    remove_articles,
    swap_words,
)



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

"""
Word-based functions for enriching English language data.

Importable functions include:

* add_hypernyms
* add_hyponyms
* add_misspelling
* add_parens
* add_synonyms
* remove_articles
* swap_words
"""

import collections
import json
from pkg_resources import resource_string
import typing

from scipy import random
from nltk import wordnet


HYPERNYMS = json.loads(resource_string("niacin", "data/hypernyms.json").decode("utf-8"))
HYPONYMS = json.loads(resource_string("niacin", "data/hyponyms.json").decode("utf-8"))
MISSPELLINGS = json.loads(
    resource_string("niacin", "data/misspellings.json").decode("utf-8")
)
SYNONYMS = json.loads(resource_string("niacin", "data/synonyms.json").decode("utf-8"))


ARTICLES = ("the", "a", "an", "these", "those", "his", "hers", "their")


def _get_wordnet():
    try:
        wn = wordnet.WordNetLemmatizer()
        wn.lemmatize("this is a test")
    except:
        print("Missing wordnet data -- attempting to download")
        import nltk

        nltk.download("wordnet")
        wn = wordnet.WordNetLemmatizer()
    return wn


def _sub_words(string: str, probability: float, mapping: typing.Mapping) -> str:
    """Replace words with a given probability.

    Split a string into words (the naïve way, on whitespace). Then, search
    word list for each key in the mapping, and replace it with its value with
    some probability. Then join them back together with a single whitespace.

    Args:
        string: text
        probability: probability of replacing a word
        mapping: map of substring -> replacement

    Returns:
        enriched text
    """
    words = string.split()
    for pattern, sub in mapping.items():
        for index, word in enumerate(words):
            if (word.lower() == pattern) and random.binomial(1, probability):
                words[index] = sub
    return " ".join(word for word in words if word)


def add_hypernyms(string: str, p: float = 0.01) -> str:
    """Replace word with a higher-level category.

    A common negative sampling technique involves replacing words
    in a sentence with a word that has the same general meaning, but
    is too general for the context, e.g.:

    "all dogs go to heaven" -> "all quadrupeds go to place"

    The replacement words are drawn from wordnet (wordnet_). For
    words with more than one possible replacement, one is selected using
    ``random.choice``.

    Args:
        string: text
        p: conditional probability of replacing a word

    Returns:
        enriched text

    .. _wordnet: https://wordnet.princeton.edu/
    """
    wn = _get_wordnet()
    words = string.split()
    lemmas = [wn.lemmatize(w) for w in words]
    for index, lemma in enumerate(lemmas):
        if (lemma in HYPERNYMS) and random.binomial(1, p):
            words[index] = random.choice(HYPERNYMS[lemma])
    return " ".join(words)


def add_hyponyms(string: str, p: float = 0.01) -> str:
    """Replace word with a lower-level category.

    A common negative sampling technique involves replacing words
    in a sentence with a word that has the same general meaning, but
    is too specific for the context, e.g.:

    "all dogs go to heaven" -> "all Australian shepherds go to heaven"

    The replacement words are drawn from wordnet (wordnet_). For
    words with more than one possible replacement, one is selected using
    ``random.choice``.

    Args:
        string: text
        p: conditional probability of replacing a word

    Returns:
        enriched text

    .. _wordnet: https://wordnet.princeton.edu/
    """
    wn = _get_wordnet()
    words = string.split()
    lemmas = [wn.lemmatize(w) for w in words]
    for index, lemma in enumerate(lemmas):
        if (lemma in HYPONYMS) and random.binomial(1, p):
            words[index] = random.choice(HYPONYMS[lemma])
    return " ".join(words)


def add_misspelling(string: str, p: float = 0.1) -> str:
    """Replace words with common misspellings.

    Replaces a word with a common way that word is mispelled, given one or
    more known, common misspellings taken from the Wikipedia spelling
    correction corpus (wikipedia_). For words with more than one common
    misspelling, one is chosen using ``random.choice``.

    Args:
        string: text
        p: conditional probability of replacing a word

    Returns:
        enriched text

    .. _wikipedia: https://en.wikipedia.org/wiki/Wikipedia:Lists_of_common_misspellings
    """
    words = string.split()
    for index, word in enumerate(words):
        if (word in MISSPELLINGS) and random.binomial(1, p):
            words[index] = random.choice(MISSPELLINGS[word])
    return " ".join(words)


def add_parens(string: str, p: float = 0.01) -> str:
    """Wrap individual words in triple parentheses.

    Adds parentheses before and after a word, e.g. ``(((term)))``.
    This is a common tactic for disrupting tokenizers and other kinds
    of word based models.

    Args:
        string: text
        p: probability of wrapping a word

    Returns:
        enriched text
    """
    words = string.split()
    for index, word in enumerate(words):
        if random.binomial(1, p):
            words[index] = "(((" + word + ")))"
    return " ".join(words)


def add_synonyms(string: str, p: float = 0.01) -> str:
    """Replace word with one that has a close meaning.

    A common data augmentation technique involves replacing words
    in a sentence with a word that has the same general meaning
    (arxiv:1509.01626_), e.g.:

    "all dogs go to heaven" -> "all domestic dog depart to heaven"

    The replacement words are drawn from wordnet (wordnet_). For
    words with more than one possible replacement, one is selected using
    ``random.choice``.

    Args:
        string: text
        p: conditional probability of replacing a word

    Returns:
        enriched text

    .. _arxiv:1509.01626 : https://arxiv.org/abs/1509.01626
    .. _wordnet: https://wordnet.princeton.edu/
    """
    wn = _get_wordnet()
    words = string.split()
    lemmas = [wn.lemmatize(w) for w in words]
    for index, lemma in enumerate(lemmas):
        if (lemma in SYNONYMS) and random.binomial(1, p):
            words[index] = random.choice(SYNONYMS[lemma])
    return " ".join(words)


def remove_articles(string: str, p: float = 1.0) -> str:
    """Remove articles from text data.

    Matches and removes the following articles:

    * the
    * a
    * an
    * these
    * those
    * his
    * hers
    * their

    with probability p.

    Args:
        string: text
        p: probability of removing a given article

    Returns:
        enriched text
    """
    mapping = {article: "" for article in ARTICLES}
    return _sub_words(string, probability=p, mapping=mapping)


def swap_words(string: str, p: float = 0.01) -> str:
    """Swap adjacent words.

    With probability p, swap two adjacent words in a string. This preserves
    the vocabulary of input text while changing token order, and in
    theory should provide more of a challenge to recursive models than ones
    that rely on lexical distributions.

    .. note::
        to keep the interface consistent, niacin's implementation acts on
        a probability p, applied n-1 times, where n is the total number of words
        in the string. In the original paper (eda_), two words are chosen n times
        and swapped, where n is a count number given as a hyperparameter.

    Args:
        string: text
        p: probability of swapping two words

    Returns:
        enriched text

    .. _eda : https://arxiv.org/abs/1901.11196
    """
    words = string.split()
    index = 0
    while index < len(words) - 1:
        if random.binomial(1, p):
            words[index], words[index + 1] = words[index + 1], words[index]
            index += 2
        else:
            index += 1
    return " ".join(words)



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

"""
Sentence-based functions for enriching English language data.

Importable functions include:

* add_applause
* add_backtranslation
* add_bytes
* add_love
"""

import regex
from scipy import random
import warnings


P_SPACE = regex.compile(r"(\s+)")


class _Translator:
    """Wrapper around fairseq language models (arXiv:1904.01038_).

    On first initialization, the instance loads language models and stores
    them as attributes on the class. New instances after this do not reload
    them. Currently implements translation from English to German, and the
    reverse.

    Attributes
    ----------
    en2de: callable
        translate from English to German
    de2en: callable
        translate from German to English
    translators: dict
        mapping of model names to model objects

    .. _arXiv:1904.01038 : https://arxiv.org/abs/1904.01038
    """

    translators: dict = {}

    def __init__(self):
        self.load_models()
        self.en2de = self.translators["en2de"].translate
        self.de2en = self.translators["de2en"].translate

    @classmethod
    def load_models(cls, force: bool = False):
        warnings.warn(
            "Backtranslation uses large translation models (~6GB) and can "
            "hours to download on the first use."
        )
        try:
            import torch
        except ImportError:
            raise ImportError(
                "torch not found - you may need to install extras with"
                "'pip install niacin[all]'"
            )
        if force or "en2de" not in cls.translators:
            cls.translators["en2de"] = torch.hub.load(
                "pytorch/fairseq",
                "transformer.wmt19.en-de.single_model",
                tokenizer="moses",
                bpe="fastbpe",
            )
        if force or "de2en" not in cls.translators:
            cls.translators["de2en"] = torch.hub.load(
                "pytorch/fairseq",
                "transformer.wmt19.de-en.single_model",
                tokenizer="moses",
                bpe="fastbpe",
            )
        # turn off dropout
        for name, model in cls.translators.items():
            model.eval()

    def backtranslate(self, string: str) -> str:
        return self.de2en(self.en2de(string))


def add_applause(string: str, p: float = 0.1) -> str:
    """Replace whitespace with clapping emojis.

    In online communities, replacing whitespace delimiters with the clapping
    emoji (\U0001f44f) is a way of indicating emphasis, possibly as a typographic
    replacement for the baton gesture. This has the unintended consequence
    of rendering word or token-based models ineffective.

    Args:
        string: text
        p: probability of replacing every whitespace character

    Returns:
        enriched text
    """
    if random.binomial(1, p):
        string = P_SPACE.sub("\U0001f44f", string)
    return string


def add_bytes(string: str, p: float = 0.1, length: int = 100) -> str:
    """Add random bytes to the end of a sentence.

    A common spam disguising technique includes appending random sequences of
    bytes to the end of text data. This can be effective against character
    based models, or loglinear models which include total length and character
    distribution as features. Random bytes are decoded as utf-8 with errors
    ignored, so the total number of characters will typically be smaller than
    the length input parameter.

    Args:
        string: text
        p: probability adding random bytes
        length: number of random bytes

    Returns:
        enriched text
    """
    if random.binomial(1, p):
        string = string + random.bytes(length).decode("utf-8", errors="replace")
    return string


def add_love(string: str, p: float = 0.1) -> str:
    """Add love to the end of a sentence.

    Appends ``' love'`` to the end of a string. Including a word with large
    positive sentiment can be used to confuse sentiment-based filters for
    input data (arXiv:1808.0911_).

    Args:
        string: text
        p: probability of adding ' love' to a sentence

    Returns:
        enriched text

    .. _arXiv:1808.0911 : https://arxiv.org/abs/1808.09115
    """
    if random.binomial(1, p):
        string = string + " love"
    return string


def add_backtranslation(string: str, p: float = 0.5) -> str:
    """Translate a sentence into another language and back.

    Use a fairseq model to translate a sentence from Enligh into German,
    then translate the German back into English with another fairseq model
    (arXiv:1904.01038_). Anecdotally, this generates sequences with similar
    semantic content, but different word choices, and is a popular way to
    augment small datasets in high resource languages (arXiv:1904.12848_).

    .. warning::
        Backtranslation uses large neural machine translation (NMT)
        models. The first time you call this function, it will download
        and cache up to 6GB of data, which can take hours depending on your
        connection speed. The slowness only happens once, but the model size
        will impact memory usage every time you use this function.

    Args:
        string: text
        p: probability of backtranslating a sentence

    Returns:
        enriched text

    .. _arXiv:1904.01038 : https://arxiv.org/abs/1904.01038
    .. _arXiv:1904.12848 : https://arxiv.org/abs/1904.12848

    """
    # the fairseq models do weird stuff with empty strings
    if not string:
        return string
    if random.binomial(1, p):
        t = _Translator()
        string = t.backtranslate(string)
    return string



#!/usr/bin/env python3
#-*- encoding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.md') as f:
    desc = f.read()

with open('requirements.txt') as f:
    install_requirements = f.read().splitlines()

extras = {}
combined = set()
for extra in ('backtranslate', 'torch'):
    with open(extra + '-requirements.txt') as f:
        requirements = f.read().splitlines()
    extras[extra] = requirements
    combined.update(requirements)
extras['all'] = sorted(combined)


setup(
    name='niacin',
    version='0.5.1',
    packages=find_packages(),
    package_data={
        'niacin': ['data/*', 'py.typed']
    },
    python_requires=">=3.6",
    install_requires=install_requirements,
    extras_require=extras,
    long_description=desc,
    long_description_content_type="text/markdown",
    tests_require=[
        'pytest',
        'pytest-cov'
    ],
)
